@@ -4,8 +4,6 @@
 #include <exception>
 #include <sstream>
 
-#include "xindex.hpp"
-
 namespace xt
 {
 
@@ -13,13 +11,13 @@ namespace xt
      * broadcast_error *
      *******************/
 
-    template <class S>
     class broadcast_error : public std::exception
     {
 
     public:
 
-        broadcast_error(const xshape<S>& lhs, const xshape<S>& rhs);
+        template <class S1, class S2>
+        broadcast_error(const S1& lhs, const S2& rhs);
 
         virtual const char* what() const noexcept;
 
@@ -32,26 +30,27 @@ namespace xt
      * broadcast_error implementation *
      **********************************/
 
-    template <class S>
-    inline broadcast_error<S>::broadcast_error(const xshape<S>& lhs,
-                                               const xshape<S>& rhs)
+    template <class S1, class S2>
+    inline broadcast_error::broadcast_error(const S1& lhs,
+                                            const S2& rhs)
     {
         std::ostringstream buf("Incompatible dimension of arrays:", std::ios_base::ate);
+        
         buf << "\n LHS shape = (";
-
-        std::ostream_iterator<S> iter1(buf, ", ");
+        using size_type1 = typename S1::value_type;
+        std::ostream_iterator<size_type1> iter1(buf, ", ");
         std::copy(lhs.begin(), lhs.end(), iter1);
 
         buf << ")\n RHS shape = (";
-        std::ostream_iterator<S> iter2(buf, ", ");
+        using size_type2 = typename S2::value_type;
+        std::ostream_iterator<size_type2> iter2(buf, ", ");
         std::copy(rhs.begin(), rhs.end(), iter2);
         buf << ")";
 
         m_message = buf.str();
     }
 
-    template <class S>
-    const char* broadcast_error<S>::what() const noexcept
+    inline const char* broadcast_error::what() const noexcept
     {
         return m_message.c_str();
     }
