@@ -73,8 +73,8 @@ namespace xt
         using size_type = std::common_type_t<typename E::size_type...>; // detail::common_size_type<E...>;
         using difference_type = std::common_type_t<typename E::difference_type...>; //detail::common_difference_type<E...>;
 
-        using shape_type = xshape<size_type>;
-        using strides_type = xstrides<size_type>;
+        using shape_type = std::vector<size_type>;
+        using strides_type = std::vector<size_type>;
         using closure_type = const self_type;
 
         using const_stepper = xfunction_stepper<F, R, E...>;
@@ -85,12 +85,18 @@ namespace xt
         xfunction(Func&& f, const E&...e) noexcept;
 
         size_type dimension() const;
+        shape_type shape() const;
 
         template <class... Args>
         const_reference operator()(Args... args) const;
 
-        bool broadcast_shape(shape_type& shape) const;
-        bool is_trivial_broadcast(const strides_type& strides) const;
+        const_reference operator[](const xindex& index) const;
+
+        template <class S>
+        bool broadcast_shape(S& shape) const;
+
+        template <class S>
+        bool is_trivial_broadcast(const S& strides) const;
 
         const_iterator begin() const;
         const_iterator end() const;
@@ -102,19 +108,23 @@ namespace xt
         const_iterator cxbegin(const shape_type& shape) const;
         const_iterator cxend(const shape_type& shape) const;
 
-        const_stepper stepper_begin(const shape_type& shape) const;
-        const_stepper stepper_end(const shape_type& shape) const;
+        template <class S>
+        const_stepper stepper_begin(const S& shape) const;
+
+        template <class S>
+        const_stepper stepper_end(const S& shape) const;
 
         const_storage_iterator storage_begin() const;
         const_storage_iterator storage_end() const;
 
-        shape_type shape() const;
-
     private:
 
         template <std::size_t... I, class... Args>
         const_reference access_impl(std::index_sequence<I...>, Args... args) const;
 
+        template <std::size_t... I>
+        const_reference access_impl(const xindex& index, std::index_sequence<I...>) const;
+
         template <class Func, std::size_t... I>
         const_stepper build_stepper(Func&& f, std::index_sequence<I...>) const;
 
@@ -195,6 +205,8 @@ namespace xt
         using difference_type = typename xfunction_type::difference_type;
         using iterator_category = std::input_iterator_tag;
 
+        using shape_type = typename xfunction_type::shape_type;
+
         template <class... It>
         xfunction_stepper(const xfunction_type* func, It&&... it);
 
@@ -277,6 +289,17 @@ namespace xt
         auto func = [](size_type d, auto&& e) { return std::max(d, e.dimension()); };
         return accumulate(func, size_type(0), m_e);
     }
+
+    /**
+     * Returns the shape of the xfunction.
+     */
+    template <class F, class R, class... E>
+    inline auto xfunction<F, R, E...>::shape() const -> shape_type
+    {
+        shape_type shape(dimension(), size_type(1));
+        broadcast_shape(shape);
+        return shape;
+    }
     //@}
 
     /**
@@ -294,6 +317,13 @@ namespace xt
     {
         return access_impl(std::make_index_sequence<sizeof...(E)>(), args...);
     }
+
+    template <class F, class R, class... E>
+    inline auto xfunction<F, R, E...>::operator[](const xindex& index) const -> const_reference
+    {
+        return access_impl(index, std::make_index_sequence<sizeof...(E)>());
+    }
+
     //@}
     
     /**
@@ -306,7 +336,8 @@ namespace xt
      * @return a boolean indicating whether the broadcast is trivial
      */
     template <class F, class R, class... E>
-    inline bool xfunction<F, R, E...>::broadcast_shape(shape_type& shape) const
+    template <class S>
+    inline bool xfunction<F, R, E...>::broadcast_shape(S& shape) const
     {
         // e.broadcast_shape must be evaluated even if b is false
         auto func = [&shape](bool b, auto&& e) { return e.broadcast_shape(shape) && b; };
@@ -319,7 +350,8 @@ namespace xt
      * @return a boolean indicating whether the broadcast is trivial
      */
     template <class F, class R, class... E>
-    inline bool xfunction<F, R, E...>::is_trivial_broadcast(const strides_type& strides) const
+    template <class S>
+    inline bool xfunction<F, R, E...>::is_trivial_broadcast(const S& strides) const
     {
         auto func = [&strides](bool b, auto&& e) { return b && e.is_trivial_broadcast(strides); };
         return accumulate(func, true, m_e);
@@ -418,14 +450,16 @@ namespace xt
     //@}
 
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::stepper_begin(const shape_type& shape) const -> const_stepper
+    template <class S>
+    inline auto xfunction<F, R, E...>::stepper_begin(const S& shape) const -> const_stepper
     {
         auto f = [&shape](const auto& e) { return e.stepper_begin(shape); };
         return build_stepper(f, std::make_index_sequence<sizeof...(E)>());
     }
 
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::stepper_end(const shape_type& shape) const -> const_stepper
+    template <class S>
+    inline auto xfunction<F, R, E...>::stepper_end(const S& shape) const -> const_stepper
     {
         auto f = [&shape](const auto& e) { return e.stepper_end(shape); };
         return build_stepper(f, std::make_index_sequence<sizeof...(E)>());
@@ -457,18 +491,6 @@ namespace xt
     }
     //@}
 
-    /**
-     * Returns the shape of the xfunction.
-     */
-    template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::shape() const -> shape_type
-    {
-        shape_type shape(dimension(), size_type(1));
-        broadcast_shape(shape);
-        return shape;
-    }
-    //@}
-
     template <class F, class R, class... E>
     template <std::size_t... I, class... Args>
     inline auto xfunction<F, R, E...>::access_impl(std::index_sequence<I...>, Args... args) const -> const_reference
@@ -476,6 +498,13 @@ namespace xt
         return m_f(detail::get_element(std::get<I>(m_e), args...)...);
     }
 
+    template <class F, class R, class... E>
+    template <std::size_t... I>
+    inline auto xfunction<F, R, E...>::access_impl(const xindex& index, std::index_sequence<I...>) const -> const_reference
+    {
+        return m_f(std::get<I>(m_e)[index]...);
+    }
+
     template <class F, class R, class... E>
     template <class Func, std::size_t... I>
     inline auto xfunction<F, R, E...>::build_stepper(Func&& f, std::index_sequence<I...>) const -> const_stepper
@@ -525,7 +554,7 @@ namespace xt
     }
 
     template <class F, class R, class... E>
-    inline bool xf_storage_iterator<F, R, E...>::equal(const xf_storage_iterator& rhs) const
+    inline bool xf_storage_iterator<F, R, E...>::equal(const self_type& rhs) const
     {
         return p_f == rhs.p_f && m_it == rhs.m_it;
     }
